#ifndef VISION_COMMON_H
#define VISION_COMMON_H
#include <ctime>
#include <chrono>
#include <Eigen/Dense>
#include <Eigen/Core>
#include <vector>
#include <iostream>
#include <ceres/ceres.h>
#include <ceres/rotation.h>

namespace Localization
{
namespace PointLoc
{
using namespace Eigen;
using namespace std;
using Pt = Eigen::Vector3d;

namespace {
  const double GRAVITY                = 9.80665;
  const double PI                     = 3.141592653589793;
  const double D_PI                   = 2 * PI;
  const double RAD_TO_DEG             = 57.295779513082323;
  const double DEG_TO_RAD             = 0.017453292519943;
  const double KMH_TO_MS              = 0.277778;    // km/h -> m/s error: 50m/s->50.00004m/s
  const double MS_TO_KMH              = 3.6;
  const double JUMP_THRE              = 5;//30m/s,20hz,一帧1.5m,给丢三帧的机会
};

class TicToc
{
public:
  TicToc()
  {
    tic();
  }

  void tic()
  {
    start_ = std::chrono::steady_clock::now();
  }

  double toc()
  {
    end_ = std::chrono::steady_clock::now();
    std::chrono::duration<double,std::milli> elapseSec = std::chrono::duration<double,std::milli>(end_ - start_);
    return elapseSec.count();
  }

private:
  std::chrono::_V2::steady_clock::time_point start_, end_;
};

struct LineCost {
  LineCost(const Eigen::Vector3d &ptlocal,const Eigen::Vector3d &ptCloet,const double &weight)
    :ptLocal_{ptlocal},
     ptClosest_{ptCloet},
     weight_{weight}
  {}

  template<typename T>
  Eigen::Matrix<T, 3, 3> GetRotRPY(const T roll,const T pitch,const T yaw) const
  {
    const Eigen::AngleAxis<T> rollAngle(Eigen::AngleAxis<T>(roll, Eigen::Matrix<T, 3, 1>::UnitX()));
    const Eigen::AngleAxis<T> pitchAngle(Eigen::AngleAxis<T>(pitch, Eigen::Matrix<T, 3, 1>::UnitY()));
    const Eigen::AngleAxis<T> yawAngle(Eigen::AngleAxis<T>(yaw, Eigen::Matrix<T, 3, 1>::UnitZ()));

    const Eigen::Quaternion<T> q = yawAngle * pitchAngle * rollAngle;
    Eigen::Matrix<T, 3, 3> R = q.toRotationMatrix();
    return R;
  }

  template<typename T>
  bool operator()(const T x[],const T y[],const T z[],
                  const T R[],const T P[],const T Y[],
                  T residual[]) const
  {
    const Eigen::Matrix<T, 3, 1> t{x[0],y[0],z[0]};
    const Eigen::Matrix<T, 3, 3> rot = GetRotRPY(R[0],P[0],Y[0]);
    Eigen::Matrix<T, 3, 1> ptTrans_ = rot * ptLocal_ + t;
    residual[0] = static_cast<T>(weight_)*(static_cast<T>(ptClosest_[0]) - static_cast<T>(ptTrans_[0]));
    residual[1] = static_cast<T>(weight_)*(static_cast<T>(ptClosest_[1]) - static_cast<T>(ptTrans_[1]));
    residual[2] = static_cast<T>(weight_)*(static_cast<T>(ptClosest_[2]) - static_cast<T>(ptTrans_[2]));
    return true;
  }
  static ceres::CostFunction *Create(const Eigen::Vector3d &p, const Eigen::Vector3d &closest_p, const float weight)
  {
    return (new ceres::AutoDiffCostFunction<
      LineCost,3,1,1,1,1,1,1>(
      new LineCost(p, closest_p, weight)));
  }
  Eigen::Vector3d ptLocal_;
  Eigen::Vector3d ptClosest_;
  double weight_;
};

static Eigen::Vector3d RotYPR2EulerYPR(const Eigen::Matrix3d &R)
{
  Eigen::Vector3d n = R.col(0);
  Eigen::Vector3d o = R.col(1);
  Eigen::Vector3d a = R.col(2);
  Eigen::Vector3d ypr;
  const double y = atan2(n(1), n(0));
  const double p = atan2(-n(2), n(0) * cos(y) + n(1) * sin(y));
  const double r = atan2(a(0) * sin(y) - a(1) * cos(y), -o(0) * sin(y) + o(1) * cos(y));
  ypr(0) = y;
  ypr(1) = p;
  ypr(2) = r;

  return ypr;
}
static Eigen::Matrix3d EulerYPR2RotYPR(const Eigen::Vector3d &ypr)
{
  double y = ypr(0);
  double p = ypr(1);
  double r = ypr(2);
  Eigen::Matrix3d mRz;
  mRz << cos(y), -sin(y), 0.0, sin(y), cos(y), 0.0, 0.0, 0.0, 1.0;

  Eigen::Matrix3d mRy;
  mRy << cos(p), 0.0, sin(p), 0.0, 1.0, 0.0, -sin(p), 0.0, cos(p);

  Eigen::Matrix3d mRx;
  mRx << 1.0, 0.0, 0.0, 0.0, cos(r), -sin(r), 0.0, sin(r), cos(r);

  return mRz * mRy * mRx;
}
///定位不用frame,也不需要多帧，slam的情况下才用frame这种结构体,如果做定位可以不用这个结构体,暂时predict功能使用此结构体
struct Frame {
  uint64_t id_{0};
  double stamp_{0.0};
  std::vector<Eigen::Vector3d> cloud_;
  Eigen::Vector3d pos_{Eigen::Vector3d::Zero()};
  Eigen::Vector3d attYPR_{Eigen::Vector3d::Zero()};
  Eigen::Quaterniond quatYPR_{Eigen::Quaterniond::Identity()};
  bool isKey_{false};
  Frame(uint64_t &id,double &stamp,std::vector<Eigen::Vector3d> &cloud) {
    id_ = id;
    stamp_ = stamp;
    cloud_ = cloud;
    pos_ = Eigen::Vector3d::Zero();
    attYPR_ = Eigen::Vector3d::Zero();
    quatYPR_ = Eigen::Quaterniond::Identity();
  }
};
} // namespace PointLoc
} // namespace Localization
#endif
